function spectra
% SPECTRA Learn about spectral analysis with MATLAB
% This function provides some supplementary material to the IOW notes on
% spectral analysis. 
%
% Info: lars.umlauf@io-warnemuende.de


%% define parameters

T        = 40;            % period
N        = 20;            % number of sampling intervals
dt       = T/N;           % sampling interval
fs       = 1/dt;          % sampling frequency
df       = 1/T;           % frequency resolution

FontSize = 12;            % for plotting


%% compute number of represented frequencies

% index of hightest representable frequency is two-sided spectra
% This depends on the way fft() deals with even and odd record numbers
Nup = N/2;
if round(Nup) == Nup 
   % even   
   Nup   = N/2 + 1;
   even  = 1;
else
    % odd
   Nup   = (N+1)/2;
   even  = 0;
end

% compute frequency vector
f        = [0:df:(Nup-1)*df]';    
omega    = 2*pi*f;


%% create some sinusoidal data

% compose time vecors
t        = [0:dt:T-dt]';

% generate input timeseries
n       = 2;                             % mode number
w1      = 2*pi*n/T;                      % frequency of model in rad/s

% create timeseries vector tracing an ellipse
Aplus   = 2;                             % counter-clockwise component                     
Aminus  = 2;                             % clockwise component                      

u       = (Aplus + Aminus)*cos(w1*t);
v       = (Aplus - Aminus)*sin(w1*t);

% complex velocity for rotary analysis
w       = u + i*v;                  

% create phase-shifted time series for correlation analysis
% (positive phase means here that u2 leads u)
phi     = pi/4;                          
u2      = (Aplus + Aminus)*cos(w1*t+phi);



%% first, compute spectra by hand

% compute FFTs of timeseries
% (everything rescaled by 1/N to the notation used in the notes)
uhat    = 1/N*fft(u);
uhat    = uhat(1:Nup);           % real spectra are one-sided

u2hat   = 1/N*fft(u2);
u2hat   = u2hat(1:Nup);     

vhat    = 1/N*fft(v);
vhat    = vhat(1:Nup);          

what    = 1/N*fft(w);            % complex spectra are two-sided

% compute real spectra
Puh     = 1/df*2*conj(uhat).*uhat;
Puh(1)  = 0.5*Puh(1);            % zero frequency component appears only once
Pu2h    = 1/df*2*conj(u2hat).*u2hat;
Pu2h(1) = 0.5*Pu2h(1);           % zero frequency component appears only once
Pvh     = 1/df*2*conj(vhat).*vhat;
Pvh(1)  = 0.5*Pvh(1);            % zero frequency component appears only once


% compute complex spectra
Pwh     = 1/df*conj(what).*what;    


%% then compute co-spectra by hand

% co-spectra of u and u2 are one-sided since both are real

% the co-spectrum
Pu12h   = 1/df*2*conj(uhat).*u2hat;   
Pu12h(1)= 0.5*Pu12h(1);                   % zero frequency component 
                                          % appears only once

% coherence
R12h    = abs(Pu12h) ./ sqrt(Puh .* Pu2h); % coherence       
R12h(Puh  < 1e-10) = 0;                    % remove spurious coherence
R12h(Pu2h < 1e-10) = 0;

% phase
Phi12h  = 180/pi*angle(Pu12h);             % phase in deg
Rratio  = R12h/max(R12h);                  % remove spurious phase shifts
Phi12h(Rratio < 1e-10) = 0;       



%% alternatively, compute spectra from Welch algorithm

% do not average, window, or overlap anything
NFFT           = N;
OverlapPercent = 0;

Hs   = spectrum.welch('Rectangular',NFFT,OverlapPercent);

% now compute the spectra
hpsd = psd(Hs,u,'NFFT',NFFT,'Fs',fs,'SpectrumType','onesided');
fu   = hpsd.Frequencies;
Pu   = hpsd.Data;

% now compute the spectra
hpsd = psd(Hs,u2,'NFFT',NFFT,'Fs',fs,'SpectrumType','onesided');
fu2  = hpsd.Frequencies;
Pu2  = hpsd.Data;

hpsd = psd(Hs,v,'NFFT',NFFT,'Fs',fs,'SpectrumType','onesided');
fv   = hpsd.Frequencies;
Pv   = hpsd.Data;

hpsd = psd(Hs,w,'NFFT',NFFT,'Fs',fs,'SpectrumType','twosided');
fw   = hpsd.Frequencies;
Pw   = hpsd.Data;


%% ...and compute co-spectra from Welch algorithm

Noverlap     = 0;
Window       = rectwin(N);
[Pu12,fu12]  = cpsd(u,u2,Window,Noverlap,NFFT,fs,'onesided');
R12          = abs(Pu12);

% coherence
R12    = abs(Pu12) ./ sqrt(Pu .* Pu2);    % coherence       
R12(Pu  < 1e-10) = 0;                     % remove spurious coherence
R12(Pu2 < 1e-10) = 0;

% phase (add 'minus' since MATLAB uses different convention)
Phi12  = - 180/pi*angle(Pu12);            % phase in deg
Rratio  = R12/max(R12);                   % remove spurious phase shifts
Phi12(Rratio < 1e-10) = 0;  


%% compute rotary spectra

% the clockwise components corresponding the right hand side of the complex
% spectra
PwhPlus      = Pwh(1:Nup);
PwPlus       = Pw(1:Nup);

% the counter-clockwise components corresonding to the left hand side
PwhMinus     = zeros(Nup,1);
PwMinus      = zeros(Nup,1);

% take into account that for even number of data points
% the hightest frequency appears only once
if even 
  PwhMinus(2:Nup-1) = Pwh(N:-1:Nup+1);
  PwMinus(2:Nup-1)  = Pw(N:-1:Nup+1);
else
  PwhMinus(2:Nup)   = Pwh(N:-1:Nup+1);
  PwMinus(2:Nup)    = Pw(N:-1:Nup+1);
end

%% check if integrated spectra and integrated variances coincide

% compute variances and compare
a = 1/T*sum(u.*u)*dt;
b = sum(Puh)*df;
c = sum(Pu)*df;
disp(' '); 

disp('Variance of u:');
disp(['From integrated variance        = ' num2str(a)]);
disp(['From integrated P (hand-made)   = ' num2str(b)]);
disp(['From integrated P (Welch)       = ' num2str(c)]);
disp(' ');

% compute co-variances and compare
a = 1/T*sum(u.*u2)*dt;
b = sum(real(Pu12h)*df);
c = sum(real(Pu12))*df;
disp(' '); 

disp('Co-variance of u and u2:');
disp(['From integrated co-variance      = ' num2str(a)]);
disp(['From integrated P (hand-made)   = ' num2str(b)]);
disp(['From integrated P (Welch)       = ' num2str(c)]);
disp(' ');


%% plotting

nFig = 0;

% input time series
nFig = nFig + 1;
figure(nFig);
clf;

subplot(2,1,1)
plot(t,u,'o-',t,v,'ro-',t,u2,'o--');
set(gca,'FontSize',FontSize)
xlabel('t')
ylabel('u')
legend('u','v','u2')
title('Input Signals');

subplot(2,1,2)
plot(u,v,'o-');
set(gca,'FontSize',FontSize)
axis equal;
xlabel('u')
ylabel('v')


% one sided real spectra
nFig = nFig + 1;
figure(nFig);
clf;

plot(f,Puh,'ko-',fu,Pu,'r*-')
set(gca,'FontSize',FontSize)
xlabel('f');
ylabel('P');
legend('hand-made','Pwelch');
title('Power-Spectral Density of u');


% correlation analysis
nFig = nFig + 1;
figure(nFig);
clf;

subplot(4,1,1)
plot(f,real(Pu12h),'ko-',fu12,real(Pu12),'r*-')
set(gca,'FontSize',FontSize,'XTickLabel',{})
title('Cross-spectral Analysis of u and u2')
ylabel('Real(P_{12})');
legend('hand-made','from cpsd');

subplot(4,1,2)
plot(f,-imag(Pu12h),'ko-',fu12,imag(Pu12),'r*-')
set(gca,'FontSize',FontSize,'XTickLabel',{})
ylabel('Imag(P_{12})');

subplot(4,1,3)
plot(f,R12h,'ko-',fu12,R12,'r*-')
set(gca,'FontSize',FontSize,'XTickLabel',{})
ylabel('|P_{12}|');

subplot(4,1,4)
plot(f,Phi12h,'ko-',fu12,Phi12,'r*-')
set(gca,'FontSize',FontSize,'YLim',[-180 180],'YTick',[-180 -90 0 90 180])
xlabel('f');
ylabel('Phase');
grid on;


% rotary spectra
nFig = nFig + 1;
figure(nFig);
clf;

subplot(2,1,1)
plot(f,PwhPlus,'ko-',f,PwhMinus,'r*--');
set(gca,'FontSize',FontSize,'XTickLabel',{})
xlabel('f');
ylabel('P');
text(0.7,0.4,'hand-made version','Units','normalized')
legend('counter-clockwise','clockwise');
title('Rotary spectra of w=u+iv');

subplot(2,1,2)
plot(f,PwPlus,'ko-',f,PwMinus,'r*--');
set(gca,'FontSize',FontSize)
xlabel('f');
ylabel('P');
text(0.7,0.4,'Welch version','Units','normalized')
legend('counter-clockwise','clockwise');





